/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: base_example.cpp,v 1.5 2002/08/20 06:31:16 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#include "base_example.h"
#include "common.h"

#include <iostream>

// $Id: base_example.cpp,v 1.5 2002/08/20 06:31:16 taku-ku Exp $;

// misc function
namespace TinySVM {

int
inc_refcount_feature_node (feature_node *f)
{
   int i;
   for (i = 0; f[i].index >= 0; i++);
   return --f[i].index;
}

int
dec_refcount_feature_node (feature_node *f)
{
   int i;
   for (i = 0; f[i].index >= 0; i++);
   return ++f[i].index;
}

int
comp_feature_node (const void *x1, const void *x2)
{
  feature_node *p1 = (feature_node *) x1;
  feature_node *p2 = (feature_node *) x2;
  return (p1->index > p2->index);
}

feature_node *
copy_feature_node (const feature_node * f)
{
  int i;
  for (i = 0; f[i].index >= 0; i++);

  try {
    feature_node *r = new feature_node[i + 1];
    for (i = 0; f[i].index >= 0; i++) {
      r[i].index = f[i].index;
      r[i].value = f[i].value;
      r[i].extra = f[i].extra->copy(); // J.K.
    }
    r[i].index = -1;
    return r;
  }

  catch (...) {
    fprintf (stderr, "copy_feature_node(): Out of memory\n");
    exit (EXIT_FAILURE);
    return 0;
  }
}

// J.K. 
// from "#tree (A (B)) #string abc" 
// to mapping {"tree" => "(A (B))", "string" => "abc"}  
void analyze_extra_part( const std::string& s, std::map< std::string, std::string >& to ) 
{
  int state = 0;
  // 0 -- init
  // 1 -- reading name
  // 2 -- reading content
  // 3 -- escape

  std::string name( 32, ' ');
  name.resize( 0 );

  std::string content( 1024, ' ');
  content.resize( 0 );

  for ( unsigned int i = 0; i < s.length(); i++ ) {
    switch ( state ) {
    case 0:
      {
        if ( s[i] == '#' ) {
          if ( name != "" ) {
            to[ name ] = content;
          }
          state = 1;
          name.resize( 0 );
        }
        break;
      }
    case 1:
      {
        if ( !isspace( s[i] ) ) {
          name.push_back( s[i] );
        } else {
          while ( i < s.length() && isspace( s[i] ) ) { i++; }
          state = 2;
          content.resize( 0 );
          if ( i < s.length() && s[i] == '#' ) { // empty part
            to[name] = content;
            name.resize( 0 );
            state = 1;
          }
          if ( i < s.length() ) { content.push_back( s[i] ); }
        }
        break;
      }
    case 2:
      {
        if ( s[i] == '#' ) {
          to[ name ] = content;
          state = 1;
          name.resize( 0 );
        } else if ( s[i] == '\\' ) {
          content.push_back( '\\' );
          state = 3;
        } else {
          content.push_back( s[i] );
        }
        break;
      }
    case 3:
      {
        content.push_back( s[i] );
        state = 2;
        break;
      }
    }
  }
  if ( name != "" ) { to[ name ] = content; }
}

feature_node *
str2feature_node (const char *s, bool ext )
{
  int elmnum = 0;
  int len = strlen (s);

  for (int i = 0; i < len; i++) {
    if (s[i] == '#') break; // J.K.
    if (s[i] == ':') elmnum++;
  }

  /*
  if ( elmnum == 0 ) {
    std::cerr << "example should contain at least one element <index>:<value>\n";
    std::cerr << std::flush;
    exit(1);
  }
  */

  try {
    feature_node *_x = new feature_node[elmnum + 1];

    int j = 0;
    int i = 0; // J.K
    for (;j < elmnum && i < len;) {
      while (i < len && isspace (s[i]))	i++;
      _x[j].index = atoi (s + i);
      while (i + 1 < len && s[i] != ':') i++;
      _x[j].value = atof (s + i + 1);
      j++;
      while (i < len && !isspace (s[i])) i++;
    }
    
    // dumy index
    _x[j].index = -1;
    _x[j].value = 0;
    if ( !ext ) {
      return _x;
    }

    // J.K.
    // find the first '#'
    while ( true ) {
      if (i == len) break;
      if (s[i] == '#') { i++; break; }
      i++;
    }
    
    // extra part
    _x->extra = new ExtraData();

    if ( i != len ) {

      std::string extra_s( s+i );
      std::map< std::string, std::string > extra_map;
      analyze_extra_part( extra_s, extra_map );

      {
        //std::cerr << "extra part:\n" << std::flush;
        //for ( std::map< std::string, std::string >::const_iterator itr = extra_map.begin(); itr != extra_map.end(); itr++ ) {
        //  std::cerr << "name=" << itr->first << " content=" << itr->second << "\n" << std::flush;
        //}
      }
      if ( extra_map.find( "no" ) != extra_map.end() ) {
        int no = atoi( extra_map[ "no" ].c_str() );
        _x->extra->setNo( no );
      }
      if ( extra_map.find( "tree" ) != extra_map.end() ) {
        // there is a tree part
        LOTree* tree = new LOTree( extra_map[ "tree" ] ); // from string
        _x->extra->setTree( tree );
      }

      if ( extra_map.find( "treev" ) != extra_map.end() ) {
        // there is a vectorized tree part
        _x->extra->setTreeVec( str2feature_node( extra_map[ "treev" ].c_str(),
                                                 false ) );
        // read with "no-extra" mode
      }
      if ( extra_map.find( "malp" ) != extra_map.end() ) {
        // there is a malicious-pair part
        std::istringstream iss( extra_map[ "malp" ] );
        iss >> std::ws;
        while ( iss ) {
          int no;
          iss >> no >> std::ws;
          _x->extra->setMaliciousPair( no );
        }
      } 
    }
   
    return _x;
  }

  catch (...) {
    fprintf (stderr, "str2feature_node(): Out of memory\n");
    exit (EXIT_FAILURE);
    return 0;
  }
}

feature_node *
fix_feature_node (feature_node * _x)
{
  register int i;
  register int cindex = -1;
  register int sorted = 1;

  // J.K.
  ExtraData* extra = _x->extra;
  _x->extra = 0;

  // check sort
  for (i = 0; _x[i].index >= 0; i++) {
     if (cindex >= _x[i].index) sorted = 0;
     cindex = _x[i].index;
  }
   
  // sort
  if (!sorted) qsort ((void *) _x, i, sizeof (feature_node), comp_feature_node);
  // J.K. restore extra field
  _x->extra = extra;

  return _x;
}

BaseExample::BaseExample ()
{
  l = d = pack_d = strl = 0;
  stre = 0;
  x = 0;
  y = 0;
  alpha = 0;
  G  = 0;
  feature_type = class_type = BINARY_FEATURE;
  feature_weight = 0;
  feature_occ    = 0;
  feature_trie   = 0;
  production_map = 0;
  max_absolute_example_no = -1;
}

BaseExample::~BaseExample ()
{
  for (int i = 0; i < l; i++) {
    if (x && dec_refcount_feature_node(x[i]) == -1) {
      if ( x[i]->extra ) { delete x[i]->extra; }// J.K.
      delete [] x[i];
    }
  }
   
  delete [] x;
  delete [] y;
  delete [] alpha;
  delete [] G;
  delete [] stre;

  //if ( feature_weight ) { delete feature_weight; }
  //if ( feature_occ ) { delete feature_occ; }
}

// copy constructor
BaseExample &
BaseExample::operator =(BaseExample & e) 
{
  if (this != &e) {
    clear ();
    for (int i = 0; i < e.l; i++) {
       inc_refcount_feature_node (e.x[i]);
       add (e.y[i], e.x[i]);
    }
    l = e.l;
    pack_d = e.pack_d;
    svindex_size = e.svindex_size;
    if (svindex_size) {
	_clone (alpha, e.alpha, svindex_size);
	_clone (G,     e.G,     svindex_size);
    }
 
    feature_weight = e.feature_weight;
    feature_occ    = e.feature_occ;
    feature_trie   = e.feature_trie;
    production_map = e.production_map;
    max_absolute_example_no = e.max_absolute_example_no;
  }

  return *this;
}

int
BaseExample::clear ()
{
  for (int i = 0; i < l; i++) {
     if (x && dec_refcount_feature_node(x[i]) == -1) { 
       if ( x[i]->extra) {delete x[i]->extra;} // J.K.
       delete [] x[i];
     }
  }

  delete [] x;
  delete [] y;
  delete [] alpha;
  delete [] G;

  l = d = pack_d = 0;
  x = 0;
  y = 0;
  alpha = 0;
  G = 0;
  return 0;
}
   

char *
BaseExample::readLine (FILE * fp)
{
  long len;
  int c;
  char *tstr;

  try {
    if (! stre) {
      strl = MAXLEN;
      stre = new char[strl];
    }

    len = 0;
    tstr = stre;

    while (1) {
      if (len >= strl) {
	tstr = _resize (tstr, strl, strl + MAXLEN, (char)0);
        strl += MAXLEN;
	stre = tstr;
      }

      c = fgetc (fp);
      if (c == '\n' || c == '\r') {
        tstr[len] = '\0';
        break;
      }

      if (c == EOF && feof (fp)) {
        tstr[len] = '\0';
        if (feof (fp) && len == 0) tstr = 0;
        break;
      }

      tstr[len++] = c;
    }

    return tstr;
  }

  catch (...) {
    fprintf (stderr, "BaseExample::readLine(): Out of memory\n");
    exit (EXIT_FAILURE);
    return 0; 
  }
}
   
int
BaseExample::remove (int i)
{
   if (i < 0 || i >= l || ! x || ! y) {
      fprintf(stderr, "BaseExample::set (): Out of range\n");
      return 0;
   }

  if (dec_refcount_feature_node(x[i]) == -1) {
    if ( x[i]->extra ) { delete x[i]->extra; }// J.K.  delete [] x[i];
  }
  for (int j = i+1; j < l; j++) {
     x[j-1] = x[j];
     y[j-1] = y[j];
  }
     
  return --l;
}

int 
BaseExample::get (int i, double &_y, feature_node *&_x)
{
   if (i < 0 || i >= l || ! x || ! y) {
      fprintf(stderr, "BaseExample::set (): Out of range\n");
      return 0;
   }
   
   _y = y[i];
   _x = x[i];
   return 1;
}

const char *
BaseExample::get (int i)
{
  if (i < 0 || i >= l || ! x || ! y) {
      fprintf(stderr, "BaseExample::get (): Out of range\n");
      return 0;
  }

  try {
    int elmnum;
    feature_node *node = x[i];
    for (elmnum = 0; node[elmnum].index >= 0; elmnum++);

    int len = _min (strl + 1024, (elmnum + 1) * 32);
    if (len > strl) {
      stre = _resize (stre, strl, len, (char)0);
      strl = len;
    }

    sprintf (stre, "%.16g", y[i]);
    char tmp[32];
    for (feature_node *node = x[i]; node->index >= 0; node++) {
      sprintf (tmp, " %d:%.16g", node->index, node->value);
      strcat(stre, tmp);
    }
    if ( node->extra ) {
      extra_stre = stre;
      extra_stre += " # ";
      extra_stre += node->extra->toString();
      return extra_stre.c_str();
    } else {
      return (const char*)stre;
    }
  }

  catch (...) {
    fprintf (stderr, "BaseExample::get (): Out of memory\n");
    exit (EXIT_FAILURE);
    return 0; 
  }
}
     
int 
BaseExample::set (int i, const double _y, feature_node *_x)
{
   if (i < 0 || i >= l || ! x || ! y) {
      fprintf(stderr, "BaseExample::set (): Out of range\n");
      return 0;
   }
   
  if (dec_refcount_feature_node(x[i]) == -1) {
    if ( x[i]->extra ) { delete x[i]->extra; }
    delete [] x[i];
  }
  _x = fix_feature_node(_x);
  inc_refcount_feature_node(_x);
   
  x[i] = _x;
  y[i] = _y;
  return 1;
}
   
int 
BaseExample::set (int i, const double _y, const char *s)
{
   return set(i, _y, (feature_node *) str2feature_node (s));
}

int 
BaseExample::set (int _i, const char *s)
{
  double _y = 0;
  int len = strlen (s);

  int i;
  for (i = 0; i < len;) {
    while (isspace (s[i])) i++;
    _y = atof (s + i);
    while (i < len && !isspace (s[i])) i++;
    while (i < len && isspace (s[i]))  i++;
    break;
  }

  return set (_i, _y, (const char *) (s + i));
}

int
BaseExample::add (const double _y, feature_node * _x)
{
  try {
    int fnum = 0;
    feature_node *node = fix_feature_node ((feature_node *) _x);
     
    // check contents
    for (int i = 0; (node + i)->index >= 0; i++) {
      if ((node + i)->value != 1) feature_type = DOUBLE_FEATURE; // check feature type
      d = _max (d, (node + i)->index);	// save max dimension
      fnum++;
    }
     
    // incriment refcount
    inc_refcount_feature_node (node);
    pack_d = _max (fnum, pack_d);

    // check class type
    if (_y != +1 && _y != -1) class_type = DOUBLE_FEATURE;

    // resize
    x = _append (x, l, node, (feature_node*)0);
    y = _append (y, l, _y, 0.0);
    l++;

    // absolute example number
    if ( node->extra && node->extra->getNo() > max_absolute_example_no ) {
      max_absolute_example_no = node->extra->getNo();
      //std::cerr << "A" << std::flush;
    }
    

    return 1;
  }

  catch (...) {
    fprintf (stderr, "BaseExample::add(): Out of memory\n");
    exit (EXIT_FAILURE);
    return 0;
  }
}

int
BaseExample::add (const double _y, const char *s)
{
  return add (_y, (feature_node *) str2feature_node (s));
}

int
BaseExample::add (const char *s)
{
  double _y = 0;
  int len = strlen (s);

  int i;
  for (i = 0; i < len;) {
    while (isspace (s[i])) i++;
    _y = atof (s + i);
    while (i < len && !isspace (s[i])) i++;
    while (i < len && isspace (s[i]))  i++;
    break;
  }

  return add (_y, (const char *) (s + i));
}

int
BaseExample::writeSVindex (const char *filename, const char *mode, const int offset)
{
  if (!alpha || !G) return 0;

  FILE *fp = fopen (filename, mode);
  if (!fp) return 0;

  for (int i = 0; i < svindex_size; i++)
    fprintf (fp, "%.16g %.16g\n", alpha[i], G[i]);

  fclose (fp);
  return 1;
}

int
BaseExample::readSVindex (const char *filename, const char *mode, const int offset)
{
  if (l == 0) {
    fprintf(stderr, "Fatal: size == 0, Read model/example file before reading .idx file\n");
    return 0;
  }

  FILE *fp = fopen (filename, mode);
  if (!fp) return 0;

  delete [] alpha;
  delete [] G;

  int _l = 0;
  char *buf;

  while ((buf = readLine (fp)) != NULL) {
    double alpha_, G_;
    if (2 != sscanf (buf, "%lf %lf\n", &alpha_, &G_)) {
      fprintf(stderr, "Fatal: Format error %s, line %d\n", filename, _l);
      fclose (fp);
      return 0;
    }

    alpha = _append (alpha, _l, alpha_, 0.0);
    G     = _append (G,     _l, G_,     0.0);
    _l++;
  }

  fclose (fp);

  //  check size of idx file
  if (l < _l) {
    fprintf(stderr, "Fatal: model/example size (%d) < idx size (%d)\n", l, _l);
    delete [] alpha;
    delete [] G;
    alpha = 0;
    G = 0;
    return 0;
  }

  svindex_size = _l;
  return 1;
}

void BaseExample::readFeatureWeightFromStream( std::istream& is )
{
  std::cerr << "Read feature weights.\n" << std::flush;

  if ( !is ) {
    std::cerr << " ... empty.\n" << std::flush;
    feature_weight 
      = new std::vector< double >( 0 );
    feature_occ  
      = new std::vector< std::vector< std::pair<unsigned int, unsigned int> > >( 0 );
    return;
  }
  std::string line;
  getline( is, line );
  int fnum = 0;
  {
    std::istringstream iss( line );
    std::string hs;
    iss >> fnum >> std::ws;
    std::cerr << " # of featuers = " << fnum << "\n" << std::flush;
  }
  feature_weight 
    = new std::vector< double >( fnum );
  feature_occ  
    = new std::vector< std::vector< std::pair<unsigned int, unsigned int> > >( fnum );
  for ( int i = 0; i < fnum; i++ ) {
    getline( is, line );
    std::istringstream iss( line );
    int no; double w; int occ_num;
    iss >> no >> w >> occ_num;
    if ( i != no ) { 
      //std::cerr << "Format error in feature weights\n" 
      //          << std::flush; 
    }
    (*feature_weight)[no] = w;
    (*feature_occ)[no].resize( occ_num );
    for( int j = 0; j < occ_num; j++ ) {
      int eno;
      int occ;
      iss >> eno >> occ;
      (*feature_occ)[no][j].first  = eno;
      (*feature_occ)[no][j].second = occ;
    }
  }
  std::cerr << "  ... end\n" << std::flush;
}

void BaseExample::readCollinsProductionMapFromStream( std::istream& is )
{
  std::cerr << "Read production map for Collins' kernel ...\n" << std::flush;
  std::string line;
  getline( is, line );
  int prod_num = 0;
  {
    std::istringstream iss( line );
    std::string hs;
    iss >> prod_num >> std::ws;
    std::cerr << " # of productions = " << prod_num << "\n" << std::flush;
  }
  
  production_map = new production_map_t();
  
  for ( int i = 0; i < prod_num; i++ ) {
    getline( is, line );
    std::istringstream iss( line );
    int no; int node_num; 
    iss >> no >> node_num >> std::ws;
    std::vector< std::pair< int, int > > v( node_num );
    v.resize( 0 );
    for ( int k = 0; k < node_num; k++ ) {
      int label; int mark;
      iss >> label >> mark >> std::ws;
      v.push_back( std::make_pair( label, mark ) );
    }
    (*production_map)[ v ] = no;
  }
  std::cerr << "  ... end\n" << std::flush;
}

void 
BaseExample::writeCollinsProductionMapToStream( std::ostream& os ) const
{
  std::cerr << "Write production map for Collins' kernel ...\n" << std::flush;
  os << production_map->size() << "\n";
  for ( production_map_t::iterator itr = production_map->begin();
        itr != production_map->end(); itr++ ) {
    const std::vector< std::pair< int, int > >& prod = itr->first;
    int no = itr->second;
    os << no << " " << prod.size();
    for ( int k = 0; k < prod.size(); k++ ) {
      os << " " << prod[k].first << " " << prod[k].second;
    }
    os << "\n";
  }
  std::cerr << "  ... end\n" << std::flush;
}

}
