/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: classifier.h,v 1.8 2002/08/20 06:31:16 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#ifndef _CLASSIFIER_H
#define _CLASSIFIER_H
#include "kernel.h"
#include "freqtm.h"


// $Id: classifier.h,v 1.8 2002/08/20 06:31:16 taku-ku Exp $;
namespace TinySVM {

class Classifier: public Kernel
{
 private:
  double model_bias;
  int    *dot_buf;
  double **binary_kernel_cache;
  int    **fi2si;

  double  (Classifier::*_getDistance) (const feature_node *) const;
  double  _getDistance_binary (const feature_node *) const;
  double  _getDistance_normal (const feature_node *) const;
  double  _getDistance_tree_restrict1( const feature_node *) const;

  //FreqtmTreeRestrict1Classification* freqtm;

 public:
  Classifier (const BaseExample &, const Param &);
  ~Classifier ();

  inline double getDistance (const char *s)
  {
    feature_node *node = str2feature_node (s);
    
    if ( (kernel_type    == TREE_COLLINS_P 
          || kernel_type == TREE_COLLINS_P_FAST)
         && node->extra && node->extra->getTree()
         && !node->extra->getTreeCollinsP() ) {
      LOTree* collp_tree 
        = node->extra->getTree()->makeCollinsPLOTree( production_map );
      node->extra->setTreeCollinsP( collp_tree );
    }
    if ( node->extra && node->extra->getTree() 
         && !node->extra->getTreeVec() && example.feature_trie 
         && !ignore_trie ) {
      // make feature vector using trie
      std::map< unsigned int, unsigned int > fmap;
      if ( kernel_type    == TREE_COLLINS_P 
           || kernel_type == TREE_COLLINS_P_FAST ) {
        // disable (this is not tested well)
        //node->extra->getTreeCollinsP()->getAllFeature( example.feature_trie, fmap );
      } else if ( kernel_type == TREE_COLLINS
                  || kernel_type == TREE_COLLINS_FAST ) {
        //node->extra->getTree()->getAllFeatureCollins( example.feature_trie, fmap );
      } else {
        //node->extra->getTree()->getAllFeature( example.feature_trie, fmap );
      }
      feature_node* fv = new feature_node[ fmap.size() + 1 ];
      int index = 0;
      for ( std::map< unsigned int, unsigned int >::const_iterator itr 
              = fmap.begin(); itr != fmap.end(); itr++ ) {
        //std::cerr << itr->first << ":" << itr->second << " " << std::flush;
        fv[ index ].index = itr->first;
        fv[ index ].value = itr->second;
        index++;
      }
      fv[ index ].index = -1;
      fv[ index ].value = 0.0;
      node->extra->setTreeVec( fv );
    }

    //std::cerr << "\n" << std::flush;
    double d = getDistance(node);
    if ( node->extra ) { delete node->extra; }// J.K.
    delete [] node;
    return d;
  }
  
  inline double getDistance (const feature_node *_x) 
  {
    return (this->*_getDistance)(_x);
  }
};
};
#endif

