/*
 This file is added by J. Kazama to the TinySVM package.
 Copyright (C) 2005. Jun'ichi Kazama <kazama@jaist.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/

#include <fstream>

#include "common.h"
#include "misc.h"
#include "example.h"
#include "base_example.h"
#include "kernel.h"
#include "param.h"

int
main (int argc, char **argv)
{
   TinySVM::Param param;

  if (!param.set (argc, argv) || argc < 3) {
    fprintf (stdout, "%s\n", COPYRIGHT);
    fprintf (stdout, "Try `kernel_test --help' for more information.\n");
    exit (EXIT_FAILURE);
  }

  fprintf (stdout, "%s\n", COPYRIGHT);

  if (param.verbose) {
    fprintf (stdout, "kernel_type:\t\t%d\n",      param.kernel_type);
    fprintf (stdout, "param_g:\t\t%g\n",          param.param_g);
    fprintf (stdout, "param_s:\t\t%g\n",          param.param_s);
    fprintf (stdout, "param_r:\t\t%g\n",          param.param_r);
    fprintf (stdout, "degree:\t\t\t%d\n",         param.degree);
  }

  TinySVM::Example example;
  if (!example.read (argv[argc - 2])) {
    fprintf (stderr, "%s: %s: No such file or directory\n", argv[0],
	     argv[argc - 2]);
    exit (EXIT_FAILURE);
  }

  if (param.verbose)
    fprintf (stdout, "feature_type:\t\t%s\n\n",
	     example.feature_type == TinySVM::BINARY_FEATURE ? "binary" : "double");

  // calculate K_ij
  TinySVM::Kernel kernel( example, param );
  double mat[ example.size() ][ example.size() ];
  double max_k  = kernel.getKernel( example.x[0], example.x[0] );
  double ave    = 0.0;
  double ld_inv = 1.0/(double)example.l;
  double denom_inv = ld_inv * ld_inv;
  std::ofstream os( argv[ argc - 1] );

  double* dest = new double[ example.l ];

  for ( int i = 0; i < example.size(); i++ ) {
    kernel.getKernel( example.x[i], example.x, example.l, dest );
    for ( int j = 0; j < example.size(); j++ ) {
      double k = dest[j];
      if ( k > max_k ) {
        max_k = k;
      }
      ave += k * denom_inv;

      mat[i][j] = k;
      os << "(" << i << ", " << j << ") = " 
         << k
         << "\n" << std::flush;
    }
  }
  os << "Maximum = " << max_k << "\n" << std::flush;
  os << "Average = " << ave   << "\n" << std::flush;

  os.close();

  // regression
  for ( int i = 0; i < example.size(); i++ ) {
    kernel.getKernel( example.x[i], example.x, example.l, dest );
    for ( int j = 0; j < example.size(); j++ ) {
      double k = dest[j];
      if ( mat[i][j] != k ) {
        std::cerr << "Regression Test Failed!!\n";
        exit(1);
      }
    }
  }
  delete[] dest;

  return EXIT_SUCCESS;
}
