/*
 This file is added by J. Kazama to the TinySVM package.
 Copyright (C) 2005. Jun'ichi Kazama <kazama@jaist.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/

#include <functional>
#include <fstream>

#include "common.h"
#include "misc.h"
#include "example.h"
#include "base_example.h"
#include "kernel.h"
#include "param.h"

struct my_less {
 public:
  bool operator()( const std::pair<int,double>& x1,
                   const std::pair<int,double>& x2 ) const {
    return std::less<double>()( x1.second, x2.second );
  }
};

int
main (int argc, char **argv)
{
   TinySVM::Param param;

  if (!param.set (argc, argv) || argc < 3) {
    fprintf (stdout, "%s\n", COPYRIGHT);
    fprintf (stdout, "Try `kernel_test --help' for more information.\n");
    exit (EXIT_FAILURE);
  }

  fprintf (stdout, "%s\n", COPYRIGHT);

  if (param.verbose) {
    fprintf (stdout, "kernel_type:\t\t%d\n",      param.kernel_type);
    fprintf (stdout, "param_g:\t\t%g\n",          param.param_g);
    fprintf (stdout, "param_s:\t\t%g\n",          param.param_s);
    fprintf (stdout, "param_r:\t\t%g\n",          param.param_r);
    fprintf (stdout, "degree:\t\t\t%d\n",         param.degree);
  }

  TinySVM::Example example;
  if (!example.read (argv[argc - 2])) {
    fprintf (stderr, "%s: %s: No such file or directory\n", argv[0],
	     argv[argc - 2]);
    exit (EXIT_FAILURE);
  }

  if (param.verbose)
    fprintf (stdout, "feature_type:\t\t%s\n\n",
	     example.feature_type == TinySVM::BINARY_FEATURE ? "binary" : "double");

  // calculate K_ij
  TinySVM::Kernel kernel( example, param );

  std::ofstream os( argv[ argc - 1] );

  std::vector< std::pair<int,double> > kk_vec( example.l );

  for ( int i = 0; i < example.size(); i++ ) {
    kk_vec[ i ].first = i;
    kk_vec[ i ].second = kernel.getKernel( example.x[i], example.x[i] );
  }

  sort( kk_vec.begin(), kk_vec.end(), my_less() );
  
  for ( int i = 0; i < example.size(); i++ ) {
    int id = kk_vec[i].first;
    os << kk_vec[i].second << "\n";
    if ( example.x[ id ]->extra
         && example.x[ id ]->extra->getTree() ) {
      std::ostringstream oss;
      oss << "tree" << id;
      os << example.x[ id ]->extra->getTree()->toString( ) << "\n";
      os << example.x[ id ]->extra->getTree()->toGraphvizString( oss.str() ) << "\n";
    }
  }

  

  os.close();
  return EXIT_SUCCESS;
}
