/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: misc.h,v 1.12 2002/08/20 06:31:17 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#ifndef _MISC_H
#define _MISC_H

#include <string>
#include <sstream>
#include <set>

#include "lotree.h"

namespace TinySVM {

  // J.K.   
  enum { LINEAR = 0, POLY = 1, NEURAL = 2, RBF = 3, ANOVA = 4, TREE = 5, TREE_ELASTIC = 6, TREE_COLLINS=7, TREE_COLLINS_PLUS=8, TREE_FAST = 9, TREE_COLLINS_FAST = 10, TREE_COLLINS_P = 11, TREE_COLLINS_P_FAST =12, TREE_ELASTIC_FAST = 13};

  enum { NO_RESTRICT = 0, RESTRICT_1 = 1, RESTRICT_2 = 2};
  enum { NO_GEOMETRIC = 0, GEOMETRIC_1 = 1, GEOMETRIC_2 = 2};
  enum { NO_THRESHOLD_NORMALIZE = 0, THRESHOLD_MALICIOUS_NORMALIZE = 1};

  enum { DOUBLE_FEATURE, BINARY_FEATURE };
  enum { SVM, SVR, ONE_CLASS };

  // J.K.
  class ExtraData;

  class feature_node
  {
  public:
    feature_node() : index( -1 ), value( 0.0 ), extra( 0 ) {}
    int    index;
    double value;
    ExtraData*  extra; 
    // J.K. extra part such as a tree
    // Only extra filed in the first node of the feature vector has meaning.
  };

  class ExtraData
  {
  private:
    int no;   // example No.
    mutable double  self_k;
    
    LOTree* tree;
    // .. add anything as you like
    
    LOTree* tree_collins_p;

    feature_node*   tree_vec;
    std::set< int > malicious_pair;

    

  public:
    ExtraData( ) : no( -1 ), self_k( -1.0 ), 
                   tree( 0 ), tree_collins_p( 0 ), tree_vec( 0 ) { 

    }
    
    ~ExtraData()
    { 
      if ( tree ) { delete tree; }
      if ( tree_vec ) { delete[] tree_vec; }
      if ( tree_collins_p ) { delete tree_collins_p; }
    }
    
    int  getNo() const { return no; }
    void setNo( int n ) { no = n; }
    
    const LOTree* getTree() const { return tree; }
    LOTree* getTree() { return tree; }
    void setTree( LOTree* _tree ) {
      tree = _tree;
    }
    
    const LOTree* getTreeCollinsP() const { return tree_collins_p; }
    LOTree* getTreeCollinsP() { return tree_collins_p; }
    void setTreeCollinsP( LOTree* _tree ) {
      tree_collins_p = _tree;
    }

    const feature_node* getTreeVec() const { return tree_vec; }
    feature_node* getTreeVec() { return tree_vec; }
    void setTreeVec( feature_node* _tree_vec ) {
      tree_vec = _tree_vec;
    }

    bool isMaliciousPair( int no ) const {
      return ( malicious_pair.find( no ) != malicious_pair.end() );
    }
    void setMaliciousPair( int no ) {
      malicious_pair.insert( no );
    }

    ExtraData* copy() {
      // implement
      std::cerr << "ExtraData::copy() not implemented." << "\n";
      exit(1);
      return 0;
    }
    
    std::string toString() const {
      std::ostringstream oss;
      oss << "#";
      if ( no >= 0 ) {
        oss << " #no " << no;
      }
      if ( tree ) {
        oss << " #tree ";
        oss << tree->toString();
      }
      if ( tree_vec ) {
        oss << " #treev";
        for (feature_node * node = tree_vec; node->index >= 0; node++) {
          oss << " " << node->index << ":" << node->value;
        }
      }
      if ( malicious_pair.size() > 0 ) {
        oss << " #malp";
        for ( std::set< int >::iterator itr = malicious_pair.begin();
              itr != malicious_pair.end(); itr++ ) {
          oss << " " << *itr;
        }
      }
      
      return oss.str();
    }
    
    void setSelfKernelSqrt( double k ) const { self_k = k; }
    double getSelfKernelSqrt() const { return self_k; }

  };
  
  
};
#endif

