/*
 TinySVM -- Yet Another Tiny SVM Package

 $Id: param.cpp,v 1.5 2002/08/20 06:31:17 taku-ku Exp $;

 Copyright (C) 2001-2002  Taku Kudo <taku-ku@is.aist-nara.ac.jp>
 All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later verjsion.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public
 License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.
*/


#include "param.h"
#include "misc.h"
#include "common.h"

#include <iostream>

// $Id: param.cpp,v 1.5 2002/08/20 06:31:17 taku-ku Exp $;
// default param

#define PARAM_MAX_SIZE 512

// split OPTION into two string to excape error in VC++ C2026
#define TINYSVM_OPTION_INFO1 "\nUsage: %s [options] training-file model-file\n\
Solver Type:\n\
  -l, --solver-type=INT              select type of solver.\n\
                                     TYPE:  0 - C-SVM (default)\n\
                                            1 - C-SVR\n\
                                            2 - One-Class-SVM (experimental)\n\
Kernel Parameter:\n\
  -t, --kernel-type=INT              select type of kernel function.\n\
                                     TYPE:  0 - linear      (w * x)  (default)\n\
                                            1 - polynomial  (s w * x + r)^d\n\
                                            2 - neural      tanh (s w * x + r)\n\
                                            3 - RBF         exp (-s * ||w-x||^2)\n\
                                            4 - ANOVA       (sum_i [exp(-s * ||w_i-x_i||^2)])^d\n\
                                            5 - tree kernel (JK)\n\
                                            6 - elastic tree kernel (JK)\n\
                                            7 - Collins (JK)\n\ 
                                            8 - Collins+ (JK)\n\
                                            9 - tree kernel (fast) (JK)\n\ 
                                           10 - Collins (fast) (JK)\n\
                                           11 - CollinsP (JK)\n\
                                           12 - CollinsP (fast) (JK)\n\
  -d, --kernel-degree=INT            set INT for parameter d in polynomial kernel. (default 1)\n\
  -r, --kernel-param-r=FLOAT         set FLOAT for parameter r in polynomial kernel. (default 1)\n\
  -s, --kernel-param-s=FLOAT         set FLOAT for parameter s in polynomial kernel. (default 1)\n\
\n"
  
  
#define TINYSVM_OPTION_INFO2 "Optimization Parameter:\n\
  -m, --cache-size=FLOAT             set FLOAT for cache memory size (MB). (default 40.0)\n\
  -c, --cost=FLOAT                   set FLOAT for cost C of constraints violation,\n\
                                     trade-off between training error and margin. (default 1.0)\n\
  -e, --termination-criterion=FLOAT  set FLOAT for tolerance of termination criterion.\n\
                                     (default 0.001)\n\
  -H, --shrinking-size=INT           set INT for number of iterations variable needs to\n\
                                     be optimal before considered for shrinking. (default 100)\n\
  -p, --shrinking-eps=FLOAT          set FLOAT for initial threshold value of shrinking process.\n\
                                     (default 2.0)\n\
  -f, --do-final-check=INT           do final optimality check for variables removed\n\
                                     by shrinking. (default 1)\n\
  -i, --insensitive-loss=FLOAT       set FLOAT for epsilon in epsilon-insensitive loss function\n\
                                     used in C-SVR cost evaluation. (default 0.1)\n\
\n\
Miscellaneous:\n\
  -M, --model=FILE                   set FILE, FILE.idx for initial condition model file.\n\
  -I, --sv-index                     write all alpha and gradient to MODEL.idx.\n\
  -W, --compress                     calculate vector w (w * x + b), instead of alpha.\n\
  -a, --geometric=INT                use geometric information in tree kernels. (JK)\n\
     TYPE: 0 - no geometric information\n\
           1 - \n\
  -b, --restriction=INT              node restriction type in tree kernels. (JK)\n\
     TYPE: 0 - no restriction\n\
           1 - sum of the products of the number of each subtree appears in the tree with the same marking. Note that the marks are treated as if they are part of the label (subtrees with different marking are treated as different).\n\
  -k, --tree-check                   check the correstness of the tree kernels by comparing with the naive implementation.\n\
  -B, --restriction-ratio=FLOAT     (1-ratio)*<not-restricted> + ratio*<restricted>\n\
  -n, --self-normalize       normalize as K(x1,x2)/sqrt(K(x1,x1)K(x2,x2)).\n\
  -z, --maximun-iteration=INT   Maximun number of iterations\n\
  -o, --auto-param-s     param_s = param_s * (1/max(_K(x,x))). \n\
  -u, --depth-factor     depth factor of tree kernel\n\
  -V, --verbose                      set verbose mode.\n\
  -v, --version                      show the version of TinySVM and exit.\n\
  -h, --help                         show this help and exit.\n\
\n"

static const char *short_options = "l:t:d:s:r:m:M:c:e:H:p:f:i:WIVvha:b:B:knou:q:z:w:E";

static struct option long_options[] = {
  {"solver-type",             required_argument, NULL, 'l'},
  {"kernel-type",             required_argument, NULL, 't'},  
  {"kernel-degree",           required_argument, NULL, 'd'},
  {"kernel-param-s",          required_argument, NULL, 's'},
  {"kernel-param-r",          required_argument, NULL, 'r'},
  {"cache-size",              required_argument, NULL, 'm'},
  {"model",                   required_argument, NULL, 'M'},
  {"cost",                    required_argument, NULL, 'c'},
  {"termination-criterion",   required_argument, NULL, 'e'},
  {"shrinking-size",          required_argument, NULL, 'H'},
  {"shrinking-eps",           required_argument, NULL, 'p'},
  {"do-final-check",          required_argument, NULL, 'f'},
  {"insensitive-loss",        required_argument, NULL, 'i'},
  {"sv-index",                no_argument,       NULL, 'I'},
  {"compress",                no_argument,       NULL, 'W'},
  {"verbose",                 no_argument,       NULL, 'V'},
  {"version",                 no_argument,       NULL, 'v'},
  {"help",                    no_argument,       NULL, 'h'},
  // J.K
  {"geometric",               required_argument, NULL, 'a'},
  {"restriction",             required_argument, NULL, 'b'},
  {"restriction-ratio",       required_argument, NULL, 'B'},
  {"tree-check",              no_argument,       NULL, 'k'},
  {"self-normalize",          no_argument,       NULL, 'n'},
  {"auto-param-s",            no_argument,       NULL, 'o'},
  {"depth-factor",            required_argument, NULL, 'u'},
  {"maximum-iteration",       required_argument, NULL, 'z'},
  {"threshold-normalize",     required_argument, NULL, 'q'},
  {"threshold-normalize-threshold", required_argument, NULL, 'w'},
  {"ignore-malicious",        no_argument, NULL, 'E'},
  {NULL, 0, NULL, 0}
};

extern char *optarg;
extern int optind;

namespace TinySVM {
  
Param::Param ()
{
  // set defualt
  kernel_type = LINEAR;
  solver_type = SVM;
  dot_kernel  = 1;
  degree = 1;
  param_g = 1;
  param_s = 1;
  param_r = 1;
  shrink_size = 100;
  shrink_eps = 2.0;
  cache_size = 40;
  insensitive_loss = 0.1;
  svindex = 0;
  C = 1;
  final_check = 1;
  eps = 0.001;
  verbose = 0;
  compress = 0;
  model[0] = 0;

  // J.K.
  tree_check            = false;
  geometric_type        = NO_GEOMETRIC;
  node_restriction_type = NO_RESTRICT;
  restriction_ratio     = 1.0;
  self_normalize        = false;
  
  auto_param_s          = false;
  depth_factor          = 1.0;
  
  maximum_iteration     = -1; // not limited    
  
  threshold_normalize            = NO_THRESHOLD_NORMALIZE;
  threshold_normalize_threshold  = 0.0;

  ignore_malicious = false;
  itrie            = false;
}

Param::~Param() {};

int
Param::set (int argc, char **argv)
{
  if (argc == 0) return 0;

  optind = 1;

  while (1) {
    int opt = getopt_long (argc, argv, short_options, long_options, NULL);
    if (opt == EOF) break;

    switch (opt) {
    case 'l':
      solver_type = atoi(optarg);
      break;
    case 't':
      kernel_type = atoi (optarg);
      if (kernel_type == RBF || kernel_type == ANOVA) dot_kernel = 0;
      break;
    case 'd':
      degree = atoi (optarg);
      break;
    case 's':
      param_s = atof (optarg);
      break;
    case 'r':
      param_r = atof (optarg);
      break;
    case 'm':
      cache_size = atof (optarg);
      break;
    case 'M':
      strcpy (model, optarg);
      break;
    case 'c':
      C = atof (optarg);
      break;
    case 'e':
      eps = atof (optarg);
      break;
    case 'H':
      shrink_size = atoi (optarg);
      break;
    case 'p':
      shrink_eps = atof (optarg);
      break;
    case 'f':
      final_check = atoi (optarg);
      break;
    case 'i':
      insensitive_loss = atof (optarg);
      break;
    case 'I':
      svindex = 1;
      break;
    case 'W':
     compress = 1;
     break;
    case 'V':
      verbose = 1;
      break;
    case 'v':
      fprintf (stdout, "%s of %s\n%s\n", argv[0], VERSION, COPYRIGHT);
      exit (EXIT_SUCCESS);
      break;
    case 'h':
      fprintf (stdout, "%s", COPYRIGHT);
      fprintf (stdout, TINYSVM_OPTION_INFO1, argv[0]);
      fprintf (stdout, TINYSVM_OPTION_INFO2);
      exit (EXIT_SUCCESS);
      break;
      // J.K.
    case 'a':
      geometric_type = atoi( optarg );
      break;
    case 'b':
      node_restriction_type = atoi ( optarg );
      break;
    case 'B':
      restriction_ratio = atof( optarg );
      break;
    case 'k':
      tree_check = true;
      break;
    case 'n':
      self_normalize  = true;
      break;
    case 'q':
      threshold_normalize = atoi( optarg );
      break;
    case 'w':
      threshold_normalize_threshold = atof( optarg );
      break;
    case 'o':
      auto_param_s    = true;
      break;
    case 'u':
      depth_factor = atof( optarg );
      break;
    case 'z':
      maximum_iteration = atoi( optarg );
      break;
    case 'E':
      ignore_malicious = true;
      break;
    default:
      return 0;
    }
  }

  if ( is_tree_kernel( kernel_type) ) {
    std::cerr << "learning with a tree kernel\n" << std::flush;
  }
  if ( is_node_lookup_kernel( kernel_type) ) {
    std::cerr << "learning with a fast node-lookup kernel\n" << std::flush;
  }

  if ( is_tree_kernel( kernel_type) ) dot_kernel = 0;
  if ( auto_param_s ) {
    param_s = -1.0;
  }
  return 1;
}

int
Param::set (const char *s)
{
  try {
    if (!s) return 1;

    char *tmp = "TinySVM::Param::set";
    char *buf = new char[strlen (s)];
    strcpy (buf, s);
    int len = strlen (buf);
    int j = 1;
    int i = 0;
    int si = 0;
    char *argv[PARAM_MAX_SIZE];
    argv[0] = tmp;

    while (i < len) {
      while (i < len && isspace (buf[i])) {
	si++;i++;  // skip beginning spaces
      };	

      while (i < len && !isspace (buf[i])) i++;	// skip contents
      buf[i++] = '\0';
      if (buf[si])
	argv[j++] = &buf[si];
      si = i;
      if (j >= PARAM_MAX_SIZE) {
	fprintf (stderr, "Param::set(): Option is too long, ignored\n");
	break;
      }
    }

    int r = set (j, argv);
    delete [] buf;
    return r;
  }

  catch (...) {
    fprintf (stderr, "Param::set(): Out of memory\n");
    exit (EXIT_FAILURE);
    return 0;
  }
}
};
